#include "game.hpp"
#include <ctime> // so i can use time()
#include <cstdlib>  // so i can use srand()
#include <iostream>
#include <iomanip>

Board::Board() {
    initialise();
}

void Board::initialise() {
    srand(time(0)); // uses the time in the system clock to create a seed for random number
    gameOver = false;
    moved = false;
    for (int i = 0; i < SIZE; ++i) { // loops through the colums and rows of the grid to create it
        for (int j = 0; j < SIZE; ++j) {
            board[i][j] = 0;
        }
    }
    generateRandomTile();
    generateRandomTile();
}

void Board::run() {
    while (!gameOver) {
        printBoard();
        userInput();
        moved = false;
    }
    std::cout << "Game Over! Thanks for playing." << std::endl;
}

void Board::generateRandomTile() {
    int emptyCells = 0;
    for (int i = 0; i < SIZE; ++i) {
        for (int j = 0; j < SIZE; ++j) {
            if (board[i][j] == 0) {  // loops through the grid to find the empty cells
                emptyCells++;  // if the cell is empty increment emptycells variable by 1
            }
        }
    }

    if (emptyCells == 0) { // check that there is an empty square to generate random box.
        gameOver = isGameOver(); // if no boxes can be created then end the game to keep inline with the games rules.
        return;
    }

    int randomIndex = std::rand() % emptyCells + 1; // this will choose a random place on the board to place a tile
    emptyCells = 0; // reset the variable
    for (int i = 0; i < SIZE; ++i) {
        for (int j = 0; j < SIZE; ++j) {
            if (board[i][j] == 0) {
                emptyCells++; 
                if (emptyCells == randomIndex) { // when the empty cells variable reaches the same number as random index it will place a tile
                    board[i][j] = (std::rand() % 2 + 1) * 2; // this randomly selects a number either 2 or 4 to set the tile to.
                    return;
                }
            }
        }
    }
}

void Board::printBoard() {
    system("clear");  // removes the terminal history
    for (int i = 0; i < SIZE; ++i) {
        for (int j = 0; j < SIZE; ++j) { // loops through the grid
            std::cout << std::setw(4) << board[i][j]; // prints the board with 4 character width between
        }
        std::cout << std::endl;
    }
}

void Board::userInput() {
    char input;
    std::cout << "Enter move (W/A/S/D): "; 
    std::cin >> input;

    switch (input) { // handles input and executes the respected movement function
        case 'W':
        case 'w':
            moveUp();
            break;
        case 'A':
        case 'a':
            moveLeft();
            break;
        case 'S':
        case 's':
            moveDown();
            break;
        case 'D':
        case 'd':
            moveRight();
            break;
        default:
            std::cout << "Invalid input. Please use W/A/S/D." << std::endl; // handles invalid move
    }

    if (!moved) {
        std::cout << "Invalid move. Try again." << std::endl;
        userInput();
    }

    generateRandomTile(); // creates a new tile if move input is accepted
}

void Board::moveLeft() {
    for (int i = 0; i < SIZE; ++i) { // loops through all the rows
        for (int j = 0; j < SIZE - 1; ++j) { // loops through the columns except the column on the right
            if (board[i][j] == 0) { // If the current tile is empty move the numbers left
                for (int k = j + 1; k < SIZE; ++k) {
                    if (board[i][k] != 0) { //checks if the position to the left is empty
                        board[i][j] = board[i][k]; // moves the current tile as far left as it can without merging
                        board[i][k] = 0; // sets the original tile to 0
                        moved = true;  // set movement bool to true to not allow 2 moves at once
                        break;
                    }
                }
            
            } else { 
                for (int k = j + 1; k < SIZE; ++k) {
                    if (board[i][k] != 0) { //check if there is a number to the left 
                        if (board[i][j] == board[i][k]) { // If Number in Grid position i j match i k 
                            board[i][j] *= 2; // then double the value of the left tile
                            board[i][k] = 0; // set the tile to the right of the merged one to 0
                            moved = true;  // set movement bool to true
                        }
                        break;
                    }
                }
            }
        }
    }
}

void Board::moveRight() {
    for (int i = 0; i < SIZE; ++i) {
        for (int j = SIZE - 1; j > 0; --j) {
            if (board[i][j] == 0) {
                for (int k = j - 1; k >= 0; --k) {
                    if (board[i][k] != 0) {
                        board[i][j] = board[i][k];
                        board[i][k] = 0;
                        moved = true; 
                        break;
                    }
                }
            } else {
                for (int k = j - 1; k >= 0; --k) {
                    if (board[i][k] != 0) {
                        if (board[i][j] == board[i][k]) {
                            board[i][j] *= 2;
                            board[i][k] = 0;
                            moved = true;  
                        }
                        break;
                    }
                }
            }
        }
    }
}

void Board::moveUp() {
    for (int j = 0; j < SIZE; ++j) {
        for (int i = 0; i < SIZE - 1; ++i) {
            if (board[i][j] == 0) {
                for (int k = i + 1; k < SIZE; ++k) {
                    if (board[k][j] != 0) {
                        board[i][j] = board[k][j];
                        board[k][j] = 0;
                        moved = true;
                        break;
                    }
                }
            } else {
                for (int k = i + 1; k < SIZE; ++k) {
                    if (board[k][j] != 0) {
                        if (board[i][j] == board[k][j]) {
                            board[i][j] *= 2;
                            board[k][j] = 0;
                            moved = true; 
                        }
                        break;
                    }
                }
            }
        }
    }
}

void Board::moveDown() {
    for (int j = 0; j < SIZE; ++j) {
        for (int i = SIZE - 1; i > 0; --i) {
            if (board[i][j] == 0) {
                for (int k = i - 1; k >= 0; --k) {
                    if (board[k][j] != 0) {
                        board[i][j] = board[k][j];
                        board[k][j] = 0;
                        moved = true;
                        break;
                    }
                }
            } else {
                for (int k = i - 1; k >= 0; --k) {
                    if (board[k][j] != 0) {
                        if (board[i][j] == board[k][j]) {
                            board[i][j] *= 2;
                            board[k][j] = 0;
                            moved = true;
                        }
                        break;
                    }
                }
            }
        }
    }
}

bool Board::isGameOver() {
    return true;
}
